#!/bin/sh
#----------------------------------------------------------------------------
# /tmp/install.sh - extlinux installation
#
# Creation   :  2021-04-23 holbru
# Last update:  $Id: install.sh 110224 2024-04-16 12:44:03Z hbfl $
#
# Copyright (c) 2021-2024 Holger Bruenjes, holgerbruenjes(at)gmx(dot)net
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#----------------------------------------------------------------------------

# set package name
package_name=extlinux

#exec 2>/tmp/extlinux-bootloader-trace$$.log
#set -x

# ---------------------------------------------------------------------------
# create extlinux bootloader
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# lockup device from /boot partition or /boot folder
# ---------------------------------------------------------------------------
lookup_device_from_path()
{
    local mountpoint="${1}"

    hex_dev_number=$(LANG=C; /usr/bin/stat ${mountpoint} |
                     ${GREP} Device: |
                     /usr/bin/cut -f 1 -d '/' |
                     /usr/bin/cut -f 2 -d ' ' |
                     /usr/bin/sed 's#h##g')

    if [ "${#hex_dev_number}" -lt "4" ]
    then
        hex_dev_number="0${hex_dev_number}"
    fi

    if [ "${#hex_dev_number}" -eq "4" ]
    then
        local major=$(printf "%d\n" "0x${hex_dev_number:0:2}")
        local minor=$(printf "%d\n" "0x${hex_dev_number:2:2}")
    else
        local major=$(printf "%d\n" "0x${hex_dev_number:0:3}")
        local minor=$(printf "%d\n" "0x${hex_dev_number:3:2}")
    fi

    for file in $(find /sys/dev/block -name "${major}:${minor}")
    do
#        source ${file}/uevent
        DEVNAME=$(${GAWK} -F'=' '/^DEVNAME/ {print $2}' "${slave}/uevent")
        echo "${DEVNAME}"
    done
}

# Trys to find the partition that /boot resides on
# This will either be on /boot or / (root)
get_boot()
{
    # Use DATA from findmnt see rc.sysint for more info
    if [ -f /proc/self/mountinfo ]
    then
        rootdev=$(findmnt -run -t noautofs -o SOURCE '/')
        bootdev=$(findmnt -run -t noautofs -o SOURCE '/boot')
    else
        echo "Could not find /proc/self/mountinfo"
        echo "Are you running a kernel greater than 2.6.24?"
        exit 1
    fi

    if [ -n "${bootdev}" ]
    then
        boot='boot'
        bootpart="${bootdev}"
#        echo "bootpart ${bootdev}"
    elif [ -n "${rootdev}" ]
    then
        boot='root'
        bootpart="${rootdev}"
#        echo "bootpart ${rootdev}"
    else
        echo "Could not find filesystem on / (root) or /boot."
        print_failed
    fi
}

# ---------------------------------------------------------------------------
# check for raid device
# ---------------------------------------------------------------------------
# return true when blockdevice is an md raid, otherwise return a unset value
# get all devices that are part of raid device $1
device_is_raid()
{
    if echo "${1}" | ${GREP} -q 'md'  &&
       [ -f /proc/mdstat ]
    then
        return 0
    else
        return 1
    fi
}

# ---------------------------------------------------------------------------
# if md devices found, get the system device name
# ---------------------------------------------------------------------------
mdraid_all_slaves()
{
    local slave slaves
    for slave in /sys/class/block/${1##*/}/slaves/*
    do
        source "${slave}/uevent"
        slaves="${slaves}/dev/${DEVNAME} "
        unset DEVNAME
    done
    echo ${slaves}
}

# ---------------------------------------------------------------------------
# check dev is partition
# ---------------------------------------------------------------------------
# Check /sys/block to see if device is partitioned
# If we have a partitioned block device (sda1) /sys/block/sda1/dev will not exist
# However, if we have an unpartitioned block device (sda) /sys/block/sda/dev will exi
dev_is_part()
{
    # $1 - blockdevice
    local dev=${1}

    # If block device uevent file should be found
    # If a partition is passed in path shouldn't exist
    if echo ${dev} | ${GREP} -q 'cciss'
    then
        if [ -f /sys/block/cciss\!${dev##*/}/dev ]
        then
             return 1
        fi
    elif echo ${dev} | ${GREP} -q 'ida'
    then
        if [ -f /sys/block/ida\!${dev##*/}/dev ]
        then
            return 1
        fi
    else
        if [ -f /sys/block/${dev##*/}/dev ]
        then
            return 1
        fi
    fi

    return 0
}

# ---------------------------------------------------------------------------
# check for gpt
# ---------------------------------------------------------------------------
# If EFI PART is present in the first 8 bytes then it must be a GPT disk
device_is_gpt()
{
    local partsig
    local tmpfile=$(/usr/bin/mktemp -t XXXXXXXXXX)
    /usr/bin/dd if="$1" skip=64 bs=8 count=1 >${tmpfile} 2>/dev/null
    read partsig < ${tmpfile}
    /usr/bin/rm -f ${tmpfile}
    if [ "${partsig}" = "EFI PART" ]
    then
        return 0
    else
        return 1
    fi
}

# ---------------------------------------------------------------------------
# clear gpt
# ---------------------------------------------------------------------------
clear_gpt_attr2()
{
    # $1 - Block Device, no partitions
    local disk=${1}

    # Special Exception for cciss controllers
    if echo ${disk} | ${GREP} -q 'cciss'
    then
        for part in /dev/cciss/${disk##*/}*p*
        do
            local partnum="${part##*[[:alpha:]]}"
            /usr/sbin/sgdisk "${disk}" --attributes="${partnum}":clear:2 &>/dev/null
        done
    # Smart 2 Controllers
    elif echo ${disk} | ${GREP} -q 'ida'
    then
        for part in /dev/ida/${disk##*/}*p*
        do
            local partnum="${part##*[[:alpha:]]}"
            /usr/sbin/sgdisk "${disk}" --attributes="${partnum}":clear:2 &>/dev/null
        done
    elif echo ${disk} | ${GREP} -q 'nvme'
    then
        for part in /dev/${disk##*/}*p*
        do
            local partnum="${part##*[[:alpha:]]}"
            /usr/sbin/sgdisk "${disk}" --attributes="${partnum}":clear:2 &>/dev/null
        done
    else
        for part in  /sys/block/${disk##*/}/${disk##*/}*
        do
            local partnum="${part##*[[:alpha:]]}"
            /usr/sbin/sgdisk "${disk}" --attributes="${partnum}":clear:2 &>/dev/null
        done
    fi
    return 0
}

# ---------------------------------------------------------------------------
# install the bootloader extlinux
# ---------------------------------------------------------------------------
install_extlinux()
{
    # set the install type
    inst_type='--install'
    if [ -f '/boot/extlinux/extlinux.conf' ]
    then
        inst_type='--update'
    fi

    if device_is_raid "${bootpart}"
    then
        echo "Detected RAID on /boot - ${inst_type##--}  EXTlinux with --raid"
        "${EXTLINUX}" ${inst_type} "${bios_bootpath}" --raid &> /dev/null
    else
        echo "${inst_type##--}  EXTlinux"
        "${EXTLINUX}" ${inst_type} "${bios_bootpath}" &> /dev/null
    fi
}

# ---------------------------------------------------------------------------
# get boot devices
# ---------------------------------------------------------------------------
# Calls get_ptb() for $bootpart or for all device in RAID
declare -A bootdevs
get_boot_devices()
{
    if device_is_raid "${bootpart}"
    then
        slaves=$(mdraid_all_slaves "${bootpart}")

        for slave in ${slaves}
        do
            if echo ${slave} | ${GREP} -Eq 'ida|cciss|nvme'
            then
                local disktmp="${slave%%[[:digit:]]}"
                local disk="${disktmp%%[p]}"
            else
                local disk="${slave%%[[:digit:]]*}"
            fi

            if device_is_gpt "${disk}"
            then
                local ptb="GPT"
            else
                local ptb="MBR"
            fi
            bootdevs="${bootdevs} "${slave}="$ptb"
        done
    else
        if echo ${bootpart} | ${GREP} -Eq 'ida|cciss|nvme'
        then
            local disktmp="${bootpart%%[[:digit:]]}"
            local disk="${disktmp%%[p]}"
        else
            local disk="${bootpart%%[[:digit:]]*}"
        fi

        if device_is_gpt "${disk}"
        then
            local ptb="GPT"
        else
            local ptb="MBR"
        fi
        bootdevs="${bootdevs} "${bootpart}="${ptb}"
    fi
}

# ---------------------------------------------------------------------------
# set active
# ---------------------------------------------------------------------------
# Function Assumes the boot partition should be marked as active
# All other partitions should not have the boot flag set
set_active()
{
    # If any bootdev is a block device without partitions bail
    # we want to set the boot flag on partitioned disk
    for dev in ${bootdevs}
    do
        dev=$(echo ${dev} | cut -d= -f1)
        if ! dev_is_part "${dev}"
        then
             echo "${dev} - is a block device. Aborting set_active!"
             print_failed
             return 1
        fi
    done

    # Clear BIOS Bootable Legacy Attribute for GPT drives
    # In rare cases where a RAID device has slaves on the same block device
    # Attribute 2 will be cleared for each partition multiple times
    for dev in ${bootdevs}
    do
        local ptb=$(echo ${dev} | cut -d= -f2)
        if [ "${ptb}" = "GPT" ]
        then
            part=$(echo ${dev} | cut -d= -f1)
            if echo ${part} | ${GREP} -Eq 'ida|cciss|nvme'
            then
                local disktmp="${part%%[[:digit:]]}"
                local disk="${disktmp%%[p]}"
            else
                local disk="${dev%%[[:digit:]]*}" #ex: /dev/sda
            fi
            clear_gpt_attr2 "${disk}"
        fi
    done

    # Set the boot flag on bootdevs (generated from get_boot_devices)
    for part in ${bootdevs}
    do
        local ptb=$(echo ${part} | cut -d= -f2)
        part=$(echo ${part} | cut -d= -f1)
        local partnum="${part##*[[:alpha:]]}"
        case "${part}" in
        *[[:digit:]]p[[:digit:]]*)
            local disk="${part%%p$partnum}" # get everything before p1
            ;;
        *)
            local disk="${part%%[[:digit:]]*}"
            ;;
        esac

        if [ "${ptb}" = "MBR" ]
        then
            if /usr/sbin/sfdisk "${disk}" --activate "${partnum}" &>/dev/null
            then
                echo "Boot Flag Set - ${part}"
            else
                echo "FAILED to Set the boot flag on ${part}"
                print_failed
                return 1
            fi
        elif [ "${ptb}" = "GPT" ]
        then
            if [ ! -e /usr/sbin/sgdisk ]
            then
                echo "FAILED to set attribute Legacy BIOS Bootable. sgdisk is not found - please install 'gptfdisk' package."
                print_failed
                return 1
            fi

            if /usr/sbin/sgdisk "${disk}" --attributes="${partnum}":set:2 &>/dev/null
            then
                echo "Attribute Legacy Bios Bootable Set - ${part}"
            else
                echo "FAILED to set attribute Legacy BIOS Bootable on ${part}"
                print_failed
                return 1
            fi
        fi
    done
    return 0
}

# ---------------------------------------------------------------------------
# install mbr
# ---------------------------------------------------------------------------
install_mbr()
{
    # If any bootdev is a block device without partitions bail
    # we want to install the mbr to a partitioned disk
    for dev in ${bootdevs}
    do
        dev=$(echo ${dev} | cut -d= -f1)
        if ! dev_is_part "${dev}"
        then
             echo "${dev} - is a block device. Aborting MBR install"
             print_failed
             return 1
        fi
    done

    for part in ${bootdevs}
    do
        local ptb=$(echo ${part} | cut -d= -f2)
        part=$(echo ${part} | cut -d= -f1)
        local partnum="${part##*[[:alpha:]]}"
        case "${part}" in
        *[[:digit:]]p[[:digit:]]*)
            local disk="${part%%p$partnum}" # get everything before p1
            ;;
        *)
            local disk="${part%%[[:digit:]]*}"
            ;;
        esac

        # We want to install to the root of the block device
        # If the device is a partition - ABORT!
        if dev_is_part "${disk}"
        then
            echo "ABORT! MBR installation to partition ($disk)!"
            print_failed
            return 1
        fi

        if [ "${ptb}" = "MBR" ]
        then
            mbrfile="${bios_libpath}/mbr.bin"
        elif [ "${ptb}" = "GPT" ]
        then
            mbrfile="${bios_libpath}/gptmbr.bin"
        fi

        if /usr/bin/dd bs=440 count=1 conv=notrunc if="${mbrfile}" of="${disk}" &> /dev/null
        then
            echo "Installed MBR (${mbrfile}) to ${disk}"
        else
            echo "Error Installing MBR (${mbrfile}) to ${disk}"
            return 1
        fi
    done
    return 0
}


# ---------------------------------------------------------------------------
# print update failed
# ---------------------------------------------------------------------------
print_failed()
{
    echo
    echo 'Setup of extlinux bootloader failed!'
    echo 'The partition on which the MBR record is to be installed'
    echo 'cannot be identified.'
    echo

    echo 'exit 0' >/usr/libexec/update-bootloader
    chmod 0750 /usr/libexec/update-bootloader
}

# ---------------------------------------------------------------------------
# main
# ---------------------------------------------------------------------------
main()
{
    bios_libpath="/tmp/extlinux"
    bios_bootpath="/boot/extlinux"
    EXTLINUX="/tmp/extlinux/extlinux"
    GREP='/usr/bin/grep'
    GAWK='/usr/bin/gawk'

    get_boot

    get_boot_devices

    set_active
    ret_ac=${?}

    if [ ${ret_ac:-1} -eq 0 ]
    then
        install_mbr
        ret=${?}

        if [ ${ret:-1} -eq 0 ]
        then
            install_extlinux

            # create the bootloader conf
            /usr/libexec/update-bootloader
        fi
    fi
    # remove install requiements
    /usr/bin/rm -rf /tmp/extlinux
    # 2024-04-16 hbfl, remove old lilo entry
    /usr/bin/rm -f /etc/lilo.conf*

    exit 0
}

# ---------------------------------------------------------------------------
# call function main
# ---------------------------------------------------------------------------
main "${@}"

# ---------------------------------------------------------------------------
# end
# ---------------------------------------------------------------------------
